/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mproclient.cpp,v 1.10 2006/04/08 19:06:10 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mproclient.cpp
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mproclient.cpp,v 1.10 2006/04/08 19:06:10 beevek Exp $

  \brief Magellan Pro client library implementation.  Connects to,
  makes requests of, and parses responses from a libmproserver-based
  server.  Note that most of these function calls block while waiting
  for a response or for an action to complete.
 */

#include "mproclient.h"
#include "protocol.h"
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <math.h>
#ifndef _WIN32
  #include <sys/time.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <netinet/tcp.h>
  #include <netdb.h>
  #include <arpa/inet.h>
  #include <unistd.h>
#else
  #include <winsock2.h>
  #include <time.h>
  #define M_PI 3.1415926535897932384626433833
#endif

#ifdef __FreeBSD__
typedef sig_t sighandler_t;
#endif
#ifdef __CYGWIN__
typedef void (*sighandler_t)(int);
#endif

extern int errno;

#define MPRO_TRANSLATE_GAIN  0.7      //! proportional gain for translate controller
#define MPRO_TRANSLATE_MIN   0.04     //! minimum translational velocity for translate controller
#define MPRO_ROTATE_GAIN     0.5      //! proportional gain for rotate controller
#define MPRO_ROTATE_MIN      0.08     //! minimum rotational velocity for rotate controller

static socket_t tcp_sock = INVALID_SOCKET;
static uint32_t id_count = 0;
static mpro_msg_t msg;
static int buf[MPRO_STATE_SIZE];
server_type_t server_type = CONTROL;

#ifndef _WIN32
// internal use: connection timeout stuff
static sighandler_t net_start_timeout(uint32_t ms, sighandler_t alrm)
{
  struct itimerval itv;
  sighandler_t old_sig;

  if(!ms)
    return SIG_DFL;
  itv.it_value.tv_sec = ms / 1000;
  itv.it_value.tv_usec = (ms % 1000) * 1000;
  itv.it_interval.tv_sec = itv.it_interval.tv_usec = 0;
  old_sig = signal(SIGALRM, (sighandler_t)alrm);
  if(setitimer(ITIMER_REAL, &itv, 0) < 0) {
    signal(SIGALRM, (sighandler_t)old_sig);
    return (sighandler_t)0xffffffff;
  }
  return old_sig;
}
static void net_stop_timeout(sighandler_t old)
{
  struct itimerval itv;
  memset(&itv, 0, sizeof(itv));
  setitimer(ITIMER_REAL, &itv, 0);
  signal(SIGALRM, old);
}
static void net_default_alarm(int sig) {}
#endif

int mpro_connect(const char *host, uint16_t port)
{
#ifdef _WIN32
  WORD version = MAKEWORD(2,2);
  WSADATA wsa_data;
  if(WSAStartup(version, &wsa_data) != 0)
    return -1;
#endif
  assert(host);
  struct sockaddr_in sa;
  sa.sin_family = AF_INET;
  sa.sin_port = htons(port);
  struct hostent *he = (struct hostent *)gethostbyname(host);
  if(!he) {
    uint32_t addr = inet_addr(host);
    if(addr == INADDR_NONE)
      return -1;
    memcpy(&sa.sin_addr, &addr, sizeof(addr));
  } else
    sa.sin_addr = *((struct in_addr *)he->h_addr);

  tcp_sock = socket(AF_INET, SOCK_STREAM, 0);
  if(tcp_sock == INVALID_SOCKET)
    return -1;
  int one = 1;
  if(setsockopt(tcp_sock, IPPROTO_TCP, TCP_NODELAY, (char *)&one, sizeof(one)) < 0) {
    mpro_disconnect();
    return -1;
  }

#ifndef _WIN32
  sighandler_t old_sig = net_start_timeout(3000, net_default_alarm);
  if(old_sig == (void *)0xffffffff)
    goto error;
#endif
  if(connect(tcp_sock, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
#ifndef _WIN32
    net_stop_timeout(old_sig);
#endif
    goto error;
  }
#ifndef _WIN32
  net_stop_timeout(old_sig);
#endif

  // get server type
  if(sock_recv(tcp_sock, (char *)&msg, sizeof(msg)) != sizeof(msg))
    goto error;
  msg.id = ntohl(msg.id);
  if(msg.id != 0 || msg.cmd != MSG_SERVER_TYPE)
    goto error;
  if(sock_recv(tcp_sock, (char *)&server_type, sizeof(server_type)) != sizeof(server_type))
    goto error;
  if(server_type != CONTROL && server_type != SIMULATOR)
    goto error;

  return 0;
 error:
  mpro_disconnect();
  return -1;
}

int mpro_disconnect(void)
{
  if(tcp_sock != INVALID_SOCKET)
    shutdown(tcp_sock,2);
  tcp_sock = INVALID_SOCKET;
#ifdef _WIN32
  WSACleanup();
#endif
  return 0;
}

// quads == number of 4-byte blocks to send
int net_send_msg_and_buf(uint8_t cmd, uint32_t quads)
{
  if(tcp_sock == INVALID_SOCKET)
    return -1;
  msg.id = htonl(++id_count);
  msg.cmd = cmd;
  if(send(tcp_sock, (char *)&msg, sizeof(msg),0) != sizeof(msg) ||
     (quads > 0 &&
      send(tcp_sock, (char *)buf, sizeof(int)*quads,0) != int(sizeof(int)*quads)))
    return -1;
  return 0;
}

// quads == number of 4-byte blocks to read
int net_read_msg_and_buf(uint32_t quads)
{
  if(tcp_sock == INVALID_SOCKET)
    return -1;
  mpro_msg_t rmsg;
  if(sock_recv(tcp_sock, (char *)&rmsg, sizeof(rmsg)) != sizeof(rmsg))
    return -1;
  // verify that this is a response to a request we've just made
  if(rmsg.id != msg.id || rmsg.cmd != msg.cmd) {
#if !defined(_WIN32) && !defined(__FreeBSD__)
    errno = EPROTO;
#endif
    return -1;
  }
  if(quads > 0 &&
     sock_recv(tcp_sock, (char *)buf, sizeof(int)*quads) != int(sizeof(int)*quads))
    return -1;
  return 0;
}

int mpro_poweroff(void)
{
  if(server_type != CONTROL)
    return -1;
  return net_send_msg_and_buf(MSG_POWEROFF, 0);
}

int mpro_set_velocity(float trans, float rot)
{
  if(fabs(trans) > MPRO_MAX_TRANS_VEL)
    trans = (trans < 0) ? -MPRO_MAX_TRANS_VEL : MPRO_MAX_TRANS_VEL;
  if(fabs(rot) > MPRO_MAX_ROT_VEL)
    rot = (rot < 0) ? -MPRO_MAX_ROT_VEL : MPRO_MAX_ROT_VEL;
  buf[0] = htonl(milli(trans));
  buf[1] = htonl(milli(rot));
  return net_send_msg_and_buf(MSG_SET_VELOCITY, 2);
}

#ifndef _WIN32
inline float max(float a, float b) { return a > b ? a : b; }
#endif

int mpro_translate(float distance)
{
  // simple proportional control
  float cur_odom = 0.0, dx, dy, tv;
  float o0[3], ot[3];
  if(mpro_get_odometry(&o0[0], &o0[1], &o0[2]) < 0)
    return -1;
  memcpy(ot, o0, 3 * sizeof(float));
  while(cur_odom < fabs(distance)) {
    dx = ot[0] - o0[0];
    dy = ot[1] - o0[1];
    cur_odom = sqrt(dx * dx + dy * dy);
    tv = MPRO_TRANSLATE_GAIN * (fabs(distance) - cur_odom) + MPRO_TRANSLATE_MIN;
    if(distance < 0)
      tv = -tv;
    if(mpro_set_velocity(tv, 0.0) < 0)
      return -1;
    if(mpro_get_odometry(&ot[0], &ot[1], &ot[2]) < 0)
      return -1;
  }
  return mpro_set_velocity(0.0, 0.0);
}

int mpro_rotate(float angle)
{
  // simple proportional control
  float cur_odom = 0.0, total_odom = 0.0, last, rv;
  float o[3];
  if(mpro_get_odometry(&o[0], &o[1], &o[2]) < 0)
    return -1;
  last = o[2];
  while(total_odom < fabs(angle)) {
    if(angle > 0)
      cur_odom = (fabs(o[2]-last) > M_PI) ? o[2]+2.0*M_PI-last : o[2]-last;
    else
      cur_odom = (fabs(o[2]-last) > M_PI) ? last-(o[2]-2.0*M_PI) : last-o[2];
    total_odom += cur_odom;
    last = o[2];
    rv = MPRO_ROTATE_GAIN * (fabs(angle) - total_odom) + MPRO_ROTATE_MIN;
    if(angle < 0)
      rv = -rv;
    if(mpro_set_velocity(0.0, rv) < 0)
      return -1;
    if(mpro_get_odometry(&o[0], &o[1], &o[2]) < 0)
      return -1;
  }
  return mpro_set_velocity(0.0, 0.0);
}

int mpro_stop(void)
{
  return mpro_set_velocity(0, 0);
}

int mpro_reset_sim(void)
{
  if(server_type != SIMULATOR)
    return -1;
  return net_send_msg_and_buf(MSG_RESET_SIM, 0);
}

int mpro_reset_odometry(float x, float y, float t)
{
  buf[0] = htonl(milli(x));
  buf[1] = htonl(milli(y));
  buf[2] = htonl(milli(t));
  return net_send_msg_and_buf(MSG_RESET_ODOMETRY, 3);
}

int mpro_get_odometry(float *x, float *y, float *t)
{
  assert(x && y && t);
  if(net_send_msg_and_buf(MSG_GET_ODOMETRY, 0) < 0 ||
     net_read_msg_and_buf(3) < 0)
    return -1;
  *x = unmilli(ntohl(buf[0]));
  *y = unmilli(ntohl(buf[1]));
  *t = unmilli(ntohl(buf[2]));
  return 0;
}

int mpro_get_all_ir(float ranges[MPRO_NUM_IR])
{
  if(net_send_msg_and_buf(MSG_GET_ALL_IR, 0) < 0 ||
     net_read_msg_and_buf(MPRO_NUM_IR) < 0)
    return -1;
  for(int i = 0; i < MPRO_NUM_IR; ++i)
    ranges[i] = unmilli(ntohl(buf[i]));
  return 0;
}

int mpro_get_ir(uint8_t index, float *range)
{
  assert(range);
  if(index >= MPRO_NUM_IR) {
    errno = EINVAL;
    return -1;
  }
  static float ranges[MPRO_NUM_IR];
  if(mpro_get_all_ir(ranges) < 0)
    return -1;
  *range = ranges[index];
  return 0;
}

int mpro_get_all_sonar(float ranges[MPRO_NUM_SONAR])
{
  if(net_send_msg_and_buf(MSG_GET_ALL_SONAR, 0) < 0 ||
     net_read_msg_and_buf(MPRO_NUM_SONAR) < 0)
    return -1;
  for(int i = 0; i < MPRO_NUM_SONAR; ++i)
    ranges[i] = unmilli(ntohl(buf[i]));
  return 0;
}

int mpro_get_sonar(uint8_t index, float *range)
{
  assert(range);
  if(index >= MPRO_NUM_SONAR) {
    errno = EINVAL;
    return -1;
  }
  static float ranges[MPRO_NUM_SONAR];
  if(mpro_get_all_sonar(ranges) < 0)
    return -1;
  *range = ranges[index];
  return 0;
}

int mpro_get_all_bump(bool toggles[MPRO_NUM_BUMP])
{
  if(net_send_msg_and_buf(MSG_GET_ALL_BUMP, 0) < 0 ||
     net_read_msg_and_buf(1) < 0)
    return -1;
  buf[0] = ntohl(buf[0]);
  for(int i = 0; i < MPRO_NUM_BUMP; ++i)
    toggles[i] = (buf[0] & (1 << i));
  return 0;
}

int mpro_get_bump(uint8_t index, bool *toggle)
{
  assert(toggle);
  if(index >= MPRO_NUM_BUMP) {
    errno = EINVAL;
    return -1;
  }
  if(net_send_msg_and_buf(MSG_GET_ALL_BUMP, 0) < 0 ||
     net_read_msg_and_buf(1) < 0)
    return -1;
  buf[0] = ntohl(buf[0]);
  *toggle = (buf[0] & (1 << index));
  return 0;
}

// not available in the header (internal use)
int mpro_get_state(int state[MPRO_STATE_SIZE])
{
  if(net_send_msg_and_buf(MSG_GET_STATE, 0) < 0 ||
     net_read_msg_and_buf(MPRO_STATE_SIZE) < 0)
    return -1;
  for(int i = 0; i < MPRO_STATE_SIZE; ++i)
    state[i] = ntohl(buf[i]);
  return 0;
}

// not available in the header (internal use)
int mpro_get_real_pose(float *x, float *y, float *t)
{
  if(server_type != SIMULATOR)
    return -1;
  assert(x && y && t);
  if(net_send_msg_and_buf(MSG_GET_REAL_POSE, 0) < 0 ||
     net_read_msg_and_buf(3) < 0)
    return -1;
  *x = unmilli(ntohl(buf[0]));
  *y = unmilli(ntohl(buf[1]));
  *t = unmilli(ntohl(buf[2]));
  return 0;
}

// not available in the header (internal use)
// note that if the server type is not SIMULATOR, this will return an
// error; if the simulator hasn't actually got a map, the pixmap
// buffer is set to null.
int mpro_get_map(uint8_t **pixbuf, uint32_t *width, uint32_t *height)
{
  if(server_type != SIMULATOR)
    return -1;
  assert(pixbuf && width && height);
  if(net_send_msg_and_buf(MSG_GET_MAP, 0) < 0 ||
     net_read_msg_and_buf(2) < 0)
    return -1;
  if(*pixbuf != 0)
    delete [] *pixbuf;
  *width = ntohl(buf[0]);
  *height = ntohl(buf[1]);
  int size = *width * *height;
  *pixbuf = new uint8_t[size];
  if(sock_recv(tcp_sock, (char *)*pixbuf, size) != size) {
    delete [] *pixbuf;
    return -1;
  }
  return 0;
}

// not available in the header (internal use)
int mpro_get_env_params(float *pixpermeter,
			float *origin_x, float *origin_y, float *origin_t)
{
  if(server_type != SIMULATOR)
    return -1;
  if(net_send_msg_and_buf(MSG_GET_ENV_PARAMS, 0) < 0 ||
     net_read_msg_and_buf(4) < 0)
    return -1;
  *pixpermeter = unmilli(ntohl(buf[0]));
  *origin_x = unmilli(ntohl(buf[1]));
  *origin_y = unmilli(ntohl(buf[2]));
  *origin_t = unmilli(ntohl(buf[3]));
  return 0;
}

#if 0
// not yet implemented
int mpro_set_l_velocity(float v);
int mpro_set_r_velocity(float v);
int mpro_set_ptz_zoom(float zoom);
int mpro_set_ptz_pantilt(float pan, float tilt);
int mpro_get_ptz_image(uint8_t **pixbuf, uint32_t *width, uint32_t *height, uint8_t *bpp);
#endif
