/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mproclient.h,v 1.4 2006/01/30 05:38:41 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mproclient.h
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mproclient.h,v 1.4 2006/01/30 05:38:41 beevek Exp $

  \brief Magellan Pro client library.  Connects to, makes requests of,
  and parses responses from a libmproserver-based server.  Note that
  most of these function calls block while waiting for a response or
  for an action to complete.
 */

#ifndef _MPROCLIENT_H
#define _MPROCLIENT_H

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus
#  define __STDC_LIMIT_MACROS
#endif
#include <inttypes.h>
#include <sys/types.h>

#define MPRO_PORT             7331    //! default network port

#define MPRO_RADIUS           0.2075  //! radius of the robot (in meters) (FIXME FIXME is this right?)

#define MPRO_NUM_BUMP         16      //! number of bump sensors

#define MPRO_NUM_IR           16      //! number of infrared range sensors
#define MPRO_IR_MIN           0.1     //! minimum range detectable by an infrared sensor
#define MPRO_IR_MAX           0.8     //! maximum range detectable by an infrared sensor

#define MPRO_NUM_SONAR        16      //! number of sonar range sensors
#define MPRO_SONAR_MIN        0.4572  //! minimum range detectable by a sonar sensor
#define MPRO_SONAR_MAX        10.0    //! maximum range detectable by a sonar sensor
#define MPRO_SONAR_BEAM_WIDTH 0.5236  //! angular beam width of the sonar sensors

#define MPRO_MAX_TRANS_VEL   0.4      //! maximum translational velocity (m/s)
#define MPRO_MAX_ROT_VEL     1.04720  //! maximum rotational velocity (rad/s)

/*!
  \brief Connect to an mpro server (generally either a simulator
  running on localhost or the robot itself)
  \param host A hostname or IP address
  \param port Port on which the server is running; the default value
  should be fine and you shouldn't have to actually pass a port.
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_connect(const char *host = "localhost", uint16_t port = MPRO_PORT);

/*!
  \brief Disconnect from the mpro server.
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_disconnect(void);

/*!
  \brief Set net translational and rotational velocities.
  \param trans Translational velocity (m/sec) -- negative values move
  the robot backward
  \param rot Rotational velocity (rad/sec) -- positive values rotate
  the robot counterclockwise, negative values clockwise
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_set_velocity(float trans = 0.0, float rot = 0.0);

/*!
  \brief Move the robot forward or backward.  Note that calling this
  procedure explicitly sets the rotational velocity of the robot to
  zero.
  \param distance Translational distance (m).  Positive values move
  the robot forward, negative values move the robot backward.
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_translate(float distance);

/*!
  \brief Rotate the robot left or right.  Note that calling this
  procedure explicitly sets the translational velocity of the robot to
  zero.
  \param angle Rotational distance (rad).  Positive values rotate the
  robot counterclockwise, negative values rotate the robot clockwise.
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_rotate(float angle);

/*!
  \brief Stop the robot.  Equivalent to setting wheel velocities to
  zero.
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_set_velocity
 */
int mpro_stop(void);

/*!
  \brief Reset the simulation to the starting state.  Note that this
  applies ONLY to a connection with a simulator.  It does nothing when
  connected to a real robot.
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_odometry
 */
int mpro_reset_sim(void);

/*!
  \brief Reset the robot's odometry counts.
  \param x x component (m)
  \param y y component (m)
  \param t theta component (rad)
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_odometry
 */
int mpro_reset_odometry(float x = 0.0, float y = 0.0, float t = 0.0);

/*!
  \brief Get the current values for the robot's odometry counts.
  \param x Pointer to location to store x component (m)
  \param y Pointer to location to store y component (m)
  \param t Pointer to location to store theta component (rad)
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_set_odometry
 */
int mpro_get_odometry(float *x, float *y, float *t);

/*!
  \brief Get the current ranges (m) reported by all IR sensors.
  \param ranges Array of length MPRO_NUM_IR into which ranges will be
  copied
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_ir
 */
int mpro_get_all_ir(float ranges[MPRO_NUM_IR]);

/*!
  \brief Get the current range (m) reported by a single IR sensor.
  \param index Index of the sensor (0 = front, counting
  counterclockwise)
  \param range Pointer to location to store range (m)
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_all_ir
 */
int mpro_get_ir(uint8_t index, float *range);

/*!
  \brief Get the current ranges (m) reported by all sonar sensors.
  \param ranges Array of length MPRO_NUM_SONAR into which ranges will
  be copied
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_sonar
 */
int mpro_get_all_sonar(float ranges[MPRO_NUM_SONAR]);

/*!
  \brief Get the current range (m) reported by a single sonar sensor.
  \param index Index of the sensor (0 = front, counting
  counterclockwise)
  \param range Pointer to location to store range (m)
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_all_sonar
 */
int mpro_get_sonar(uint8_t index, float *range);

/*!
  \brief Get the current toggle values (1=on, 0=off) reported by all
  bump sensors.
  \param toggles Array of length MPRO_NUM_BUMP into which toggle
  values will be copied
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_bump
 */
int mpro_get_all_bump(bool toggles[MPRO_NUM_BUMP]);

/*!
  \brief Get the current toggle value (1=on, 0=off) reported by a
  single bump sensor.
  \param index Index of the sensor (0 = front, counting
  counterclockwise)
  \param toggle Pointer to location to store toggle value
  \return >= 0 on success; < 0 on failure (check errno)
  \sa mpro_get_all_bump
 */
int mpro_get_bump(uint8_t index, bool *toggle);


// not yet implemented
#if 0
/*!
  \brief Set left wheel velocity.
  \param v Velocity (revolutions/sec)
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_set_l_velocity(float v = 0.0);

/*!
  \brief Set right wheel velocity.
  \param v Velocity (revolutions/sec)
  \return >= 0 on success; < 0 on failure (check errno)
 */
int mpro_set_r_velocity(float v = 0.0);

int mpro_set_ptz_zoom(float zoom = 0.0); // 0..1
int mpro_set_ptz_pantilt(float pan = 0.0, float tilt = 0.0); // -rad .. +rad
int mpro_get_ptz_image(uint8_t **pixbuf, uint32_t *width, uint32_t *height, uint8_t *bpp);
#endif

#ifdef __cplusplus
}
#endif

#endif // _MPROCLIENT_H
