/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mprod.cpp,v 1.6 2006/02/08 02:24:55 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mprod.cpp
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mprod.cpp,v 1.6 2006/02/08 02:24:55 beevek Exp $

  \brief Magellan Pro server daemon.  Runs onboard the robot and
  listens for connections from libmproclient, and then translates
  commands from clients into MAGE calls.
 */

#include "mproserver.h"
#include "options.hpp"
#include <string>
#include <iostream>
#include <fstream>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <Mage/Mage.h>

long int mage_state[NUM_STATE];

/*! \brief Signal handler, shut down gracefully */
void mprod_shutdown(int sig)
{
  if(sig > 0)
    printf("Received signal %d, shutting down...", sig);
  else
    printf("Shutting down...");
  mpro_server_stop();
  sonarOff();
  irOff();
  usleep(2000); // for some reason this fixes problems in disconnectRobot
  disconnectRobot();
  printf("done.\n");
  signal(sig, SIG_DFL);
  kill(getpid(), sig);
}

int main(int argc, char **argv)
{
  // set up commandline/config file options
  options::add<bool>("help", 0, "Print usage information", 0, false, options::nodump);
  options::set_cf_options("config", "c");
  options::add<std::string>("save-config", 0, "Save configuration file", 0, "", options::nodump);
  options::add<std::string>("logfile", "l", "Path to logfile (must be writable)", 0, "");
  options::add<bool>("foreground", "f", "Foreground: do not run as a daemon", 0, false);

  // parse commandline (and read config file if it is specified)
  int inpidx = options::parse_cmdline(argc, argv);
  if(inpidx < 0) // some kind of error
    return 1;    // relevant info printed by getopt

  // print usage information
  if(inpidx > argc || options::quickget<bool>("help") == true) {
    std::cerr << "Usage: " << argv[0] << " [options]" << std::endl;
    options::print_options(std::cout);
    return 1;
  }

  // save a config file based on these options?
  std::string cfname = options::quickget<std::string>("save-config");
  if(cfname.length() > 0) {
    std::ofstream conf(cfname.c_str());
    if(!conf)
      std::cerr << "Can't write configuration file " << cfname << std::endl;
    else
      options::dump(conf);
  }

  if(!options::quickget<bool>("foreground")) {
    if(daemon(1, 1) < 0) {
      perror("daemon()");
      return 1;
    }
    const char *log = "/dev/null";
    if(options::quickget<std::string>("logfile") != "")
      log = options::quickget<std::string>("logfile").c_str();
    fclose(stdin);
    freopen(log, "a", stdout);
    freopen(log, "a", stderr);
  }

  signal(SIGINT, mprod_shutdown);
  signal(SIGQUIT, mprod_shutdown);
  signal(SIGTERM, mprod_shutdown);
  signal(SIGSEGV, mprod_shutdown); // hopefully never!

  connectRobot(mage_state);
  irOn();
  sonarOn();

  if(mpro_server_start(CONTROL) < 0 || mpro_serve_forever() < 0)
    perror("ERROR");

  // should never get here
  mprod_shutdown(0);
  return 1;
}

/*
  Implementations of client request handlers.
*/

int mpro_poweroff(void)
{
  mpro_server_stop();
  sonarOff();
  irOff();
  disconnectRobot();
  if(execl("/sbin/halt", "halt", "-p", (char *)NULL) < 0)
    return -1;
  exit(0);
  return 0; // never
}

int mpro_set_velocity(float trans, float rot)
{
  if(fabs(trans) > MPRO_MAX_TRANS_VEL)
    trans = (trans < 0) ? -MPRO_MAX_TRANS_VEL : MPRO_MAX_TRANS_VEL;
  if(fabs(rot) > MPRO_MAX_ROT_VEL)
    rot = (rot < 0) ? -MPRO_MAX_ROT_VEL : MPRO_MAX_ROT_VEL;
  return vm(milli(trans), milli(rot));
}

int mpro_stop(void)
{
  return mpro_set_velocity(0, 0);
}

int mpro_reset_odometry(float x, float y, float t)
{
  return jump(milli(x), milli(y), milli(t));
}

int mpro_get_odometry(float *x, float *y, float *t)
{
  assert(x != 0 && y != 0 && t != 0);
  *x = unmilli(mage_state[STATE_X]);
  *y = unmilli(mage_state[STATE_Y]);
  *t = unmilli(mage_state[STATE_T]);
  return 0;
}

int mpro_get_all_ir(float ranges[MPRO_NUM_IR])
{
  for(int i = 0; i < MPRO_NUM_IR; ++i)
    ranges[i] = unmilli(mage_state[STATE_IR_0+i]);
  return 0;
}

int mpro_get_ir(uint8_t index, float *range)
{
  assert(index < MPRO_NUM_IR && range != 0);
  *range = unmilli(mage_state[STATE_IR_0+index]);
  return 0;
}

int mpro_get_all_sonar(float ranges[MPRO_NUM_SONAR])
{
  for(int i = 0; i < MPRO_NUM_SONAR; ++i)
    ranges[i] = unmilli(mage_state[STATE_SONAR_0+i]);
  return 0;
}

int mpro_get_sonar(uint8_t index, float *range)
{
  assert(index < MPRO_NUM_SONAR && range != 0);
  *range = unmilli(mage_state[STATE_SONAR_0+index]);
  return 0;
}

int mpro_get_all_bump(bool toggles[MPRO_NUM_BUMP])
{
  for(int i = 0; i < MPRO_NUM_BUMP; ++i)
    toggles[i] = (mage_state[STATE_BUMPER] >> i) & 0x1;
  return 0;
}

int mpro_get_bump(uint8_t index, bool *toggle)
{
  assert(index < MPRO_NUM_BUMP && toggle != 0);
  *toggle = (mage_state[STATE_BUMPER] >> index) & 0x1;
  return 0;
}

int mpro_get_state(int state[MPRO_STATE_SIZE])
{
  memcpy(state, mage_state, MPRO_STATE_SIZE * sizeof(int));
  return 0;
}

// not yet implemented
int mpro_set_l_velocity(float v) { return -1; }
int mpro_set_r_velocity(float v) { return -1; }
int mpro_set_ptz_zoom(float zoom) { return -1; }
int mpro_set_ptz_pantilt(float pan, float tilt) { return -1; }
int mpro_get_ptz_image(uint8_t **pixbuf, uint32_t *width, uint32_t *height, uint8_t *bpp) { return -1; }

// only on the simulator
int mpro_reset_sim(void)
{
  return -1;
}
int mpro_get_real_pose(float *x, float *y, float *t)
{
  return -1;
}
int mpro_get_map(uint8_t **pixbuf, uint32_t *width, uint32_t *height)
{
  return -1;
}
int mpro_get_env_params(float *pixpermeter,
			float *origin_x, float *origin_y, float *origin_t)
{
  return -1;
}
