/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mprogui.cpp,v 1.5 2006/02/09 01:42:20 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mprogui.cpp
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mprogui.cpp,v 1.5 2006/02/09 01:42:20 beevek Exp $

  \brief OpenGL based GUI.  Acts as a client (based on libmproclient)
  and displays state information about the robot.  If connected to a
  simulator, also displays the simulated environment and the actual
  robot configuration.
 */

#include "mproclient.h"
#include "protocol.h"
#include "options.hpp"
#include <stdio.h>
#include <string>
#include <iostream>
#include <fstream>
#include <GL/glut.h>
#include <vector>
#include <math.h>
#ifndef _WIN32
  #include <unistd.h>
#else
  #define M_PI 3.1415926535897932384626433833
#endif

// from MAGE:
#define STATE_IR_0                 1
#define STATE_SONAR_0              17 
#define STATE_BUMPER               33
#define STATE_X                    34
#define STATE_Y                    35
#define STATE_T                    36
#define STATE_VEL_TRANS            37
#define STATE_VEL_STEER            38

// from libmproclient, internal use only:
extern server_type_t server_type;
extern int mpro_get_state(int state[MPRO_STATE_SIZE]);
extern int mpro_get_real_pose(float *x, float *y, float *t);
extern int mpro_get_map(uint8_t **pixbuf, uint32_t *width, uint32_t *height);
extern int mpro_get_env_params(float *pixpermeter,
			       float *origin_x, float *origin_y, float *origin_t);

struct pose_t
{
  float x, y, t;
};

int mage_state[MPRO_STATE_SIZE]; // state vector in MAGE format
pose_t real_pose;
std::vector<pose_t> trajectory, real_trajectory;
bool show_real_pose = false, show_map = false, show_trajectory = true;

uint8_t *map = 0; // map pixmap (used only if the server is a simulator)
uint32_t map_w, map_h;
uint32_t gui_w = 800, gui_h = 800, orig_gui_w, orig_gui_h;
float workspace_w, workspace_h, extra_x = 0, extra_y = 0;
float pixpermeter = 80.0;
float origin[] = {0.0, 0.0, 0.0};

// some gui state
float scale = 1.0;
int mouse_x = 0, mouse_y = 0;
bool mouse_left = false, mouse_right = false;
float trans_x = 0.0, trans_y = 0.0;


// render a robot at the specified pose, along with its trajectory so
// far and its sensor states
void draw_robot(float x, float y, float t,
		const std::vector<pose_t> &traj, float alpha = 1.0)
{
  if(show_trajectory) { // trajectory
    glLineWidth(1);
    glColor4f(0.0, 0.8, 0.0, alpha);
    glBegin(GL_LINE_STRIP);
    for(int i = 0; i < int(traj.size()); ++i)
      glVertex2d(traj[i].x, traj[i].y);
    glEnd();
  }

  // switch to a robot-centric coordinate system
  glPushMatrix();
  glLoadIdentity();
  glTranslatef(x, y, 0);
  glRotatef(t * 180.0 / M_PI, 0, 0, 1);

  // chassis
  glColor4f(0.0, 0.4, 0.0, alpha);
  glBegin(GL_TRIANGLE_FAN);
  float a = M_PI/16.0;
  for(int i = 0; i <= MPRO_NUM_IR; ++i, a += M_PI/8.0)
    glVertex2d(MPRO_RADIUS * cos(a), MPRO_RADIUS * sin(a));
  glEnd();

  // draw an arrow to show orientation
  glColor4f(0.0, 1.0, 0.0, alpha);
  glBegin(GL_LINES);
  glVertex2d(0, 0);
  glVertex2d(2.0 * MPRO_RADIUS / 3.0, 0);
  glEnd();
  glBegin(GL_LINES);
  glVertex2d(2.0 * MPRO_RADIUS / 3.0, 0);
  glVertex2d(MPRO_RADIUS / 2.0, MPRO_RADIUS / 5.0);
  glEnd();
  glBegin(GL_LINES);
  glVertex2d(2.0 * MPRO_RADIUS / 3.0, 0);
  glVertex2d(MPRO_RADIUS / 2.0, -MPRO_RADIUS / 5.0);
  glEnd();

  float range;
  // sonar
  glColor4f(0.5, 0.7, 1.0, 0.7 * alpha);
  for(int i = 0; i < MPRO_NUM_SONAR; ++i) {
    range = unmilli(mage_state[STATE_SONAR_0+i]);
    if(range > MPRO_SONAR_MAX)
      range = MPRO_SONAR_MAX;
    a = float(i) * 360.0/float(MPRO_NUM_SONAR);
    glPushMatrix();
    glRotatef(a, 0, 0, 1);
    glTranslatef(MPRO_RADIUS, 0.0, 0.0);
    glBegin(GL_TRIANGLE_FAN);
    glVertex2d(0.0, 0.0);
    float sa = -MPRO_SONAR_BEAM_WIDTH / 2.0;
    for(int j = 0; j < 15; ++j, sa += MPRO_SONAR_BEAM_WIDTH / float(15))
      glVertex2d(range * cos(sa), range * sin(sa));
    glEnd();
    glPopMatrix();
  }

  // ir
  glColor4f(1.0, 0.0, 0.0, 0.9 * alpha);
  for(int i = 0; i < MPRO_NUM_IR; ++i) {
    range = unmilli(mage_state[STATE_IR_0+i]);
    if(range > MPRO_IR_MAX)
      range = MPRO_IR_MAX;
    a = float(i) * 2.0 * M_PI/float(MPRO_NUM_IR);
    glBegin(GL_LINES);
    glVertex2d(MPRO_RADIUS * cos(a), MPRO_RADIUS * sin(a));
    glVertex2d((MPRO_RADIUS + range) * cos(a), (MPRO_RADIUS + range) * sin(a));
    glEnd();
  }

  // bump
  a = -M_PI/16.0;
  for(int i = 0; i < MPRO_NUM_BUMP; ++i, a += M_PI/8.0) {
    glColor4f(0.0, (mage_state[STATE_BUMPER] >> i) & 0x1 ? 1.0 : 0.0, 0.0, alpha);
    glBegin(GL_QUADS);
    glVertex2d(MPRO_RADIUS * cos(a), MPRO_RADIUS * sin(a));
    glVertex2d(MPRO_RADIUS * cos(a+M_PI/8.0), MPRO_RADIUS * sin(a+M_PI/8.0));
    glVertex2d((MPRO_RADIUS+0.05) * cos(a+M_PI/8.0), (MPRO_RADIUS+0.05) * sin(a+M_PI/8.0));
    glVertex2d((MPRO_RADIUS+0.05) * cos(a), (MPRO_RADIUS+0.05) * sin(a));
    glEnd();
  }

  glPopMatrix();
}

void render_map()
{
  if(map == 0)
    return;

  // if the raster position is offscreen, we still want to draw the
  // bitmap, so we use a trick from teh intarnet!

  // find the raster position
  GLdouble mview[16], proj[16], x, y, z;
  GLint vport[4];
  glGetDoublev(GL_MODELVIEW_MATRIX, mview);
  glGetDoublev(GL_PROJECTION_MATRIX, proj);
  glGetIntegerv(GL_VIEWPORT, vport);
  gluProject(-workspace_w / 2, -workspace_h / 2, 0.0, mview, proj, vport, &x, &y, &z);

  // save state
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  // trick!
  gluOrtho2D(0, gui_w, 0, gui_h);
  glRasterPos2d(0,0);
  uint8_t b[1] = {0};
  glBitmap(1, 1, 0.0, 0.0, x, y, b);

  // restore state
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  // render pixels
  glPixelStorei(GL_UNPACK_ROW_LENGTH, map_w);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glPixelZoom(pixpermeter * workspace_w / scale / float(map_w),
 	      pixpermeter * workspace_h / scale / float(map_h));
  glDrawPixels(map_w, map_h, GL_LUMINANCE, GL_UNSIGNED_BYTE, map);
}

void gui_display()
{
  glClearColor(1, 1, 1, 1);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(scale * (-workspace_w / 2 + trans_x),
	     scale * (workspace_w / 2 + trans_x + extra_x),
 	     scale * (-workspace_h / 2 + trans_y),
	     scale * (workspace_h / 2 + trans_y + extra_y));

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if(show_map)
    render_map();

  if(show_real_pose) // render robot/trajectory/sensors according to true pose
    draw_robot(real_pose.x, real_pose.y, real_pose.t, real_trajectory, 0.4);

  // render robot/trajectory/sensors according to odometry
  draw_robot(unmilli(mage_state[STATE_X]) + origin[0],
	     unmilli(mage_state[STATE_Y]) + origin[1],
	     unmilli(mage_state[STATE_T]) + origin[2],
	     trajectory, 1.0);

  glutSwapBuffers();
}

void gui_reshape(int w, int h)
{
  gui_w = w;
  gui_h = h;
  glViewport(0, 0, w, h);
  extra_x = workspace_w * (float(w) - orig_gui_w) / float(orig_gui_w);
  extra_y = workspace_h * (float(h) - orig_gui_h) / float(orig_gui_h);
}

void gui_keyboard(uint8_t key, int x, int y)
{
  if(key == 'Q' || key == 'q') { // quit
    printf("Ciao!\n");
    mpro_disconnect();
    exit(0);
  } else if(key == 'H' || key == 'h') { // print help
    printf("Mouse commands:\n"
	   " Left+drag: pan the scene\n"
	   " Right+drag: zoom the scene\n\n");
    printf("Keyboard commands:\n"
	   " HOME: reset zoom/pan\n"
	   " q: quit\n"
	   " h: help\n"
	   " t: show/hide robot trajectory\n");
    if(server_type == SIMULATOR) {
      printf(" p: show/hide real robot pose\n"
	     " m: show/hide map\n");
    }
  } else if(key == 'T' || key == 't') { // show trajectory
    show_trajectory = !show_trajectory;
    printf(show_trajectory ? "Showing the robot's trajectory\n" :
	   "Hiding the robot's trajectory\n");
  }

  if(server_type == SIMULATOR) {
    switch(key) {
    case 'P': // show real pose
    case 'p':
      show_real_pose = !show_real_pose;
      printf(show_real_pose ? "Showing the real robot pose\n" :
	     "Hiding the real robot pose\n");
      break;

    case 'M': // show map
    case 'm':
      show_map = !show_map;
      printf(show_map ? "Showing the map\n" : "Hiding the map\n");
      break;
    }
  }

  glutPostRedisplay();
}

void gui_special(int key, int x, int y)
{
  if(key == GLUT_KEY_HOME) {
    scale = 1.0;
    trans_x = trans_y = 0.0;
  }
  glutPostRedisplay();
}

void gui_mouse(int button, int state, int x, int y)
{
  mouse_x = x;
  mouse_y = y;

  if(state == GLUT_UP) {
    if(button == GLUT_LEFT_BUTTON) mouse_left = false;
    else if(button == GLUT_RIGHT_BUTTON) mouse_right = false;
  } else {
    if(button == GLUT_LEFT_BUTTON) mouse_left = true;
    else if(button == GLUT_RIGHT_BUTTON) mouse_right = true;
  }
}

void gui_motion(int x, int y)
{
  int dx = x - mouse_x, dy = y - mouse_y;
  if(dx == 0 && dy == 0)
    return;

  if(mouse_left) { // pan
    trans_x -= workspace_w * float(dx) / float(orig_gui_w);
    trans_y += workspace_h * float(dy) / float(orig_gui_h);
  } else if(mouse_right) // zoom
    scale *= exp(0.05 * double(dy) + 0.01);

  mouse_x = x;
  mouse_y = y;

  if(mouse_left || mouse_right)
    glutPostRedisplay();
}

void gui_idle()
{
  static int tcount = 1; // count for storing trajectories
  static pose_t tpose;
  if(--tcount < 0) tcount = 20;

  // sleep for a while to keep down the network traffic
  //  usleep(25000);

  // get an update of the robot state
  if(mpro_get_state(mage_state) < 0) {
    perror("I can't get the robot state!");
#ifndef _WIN32
    usleep(25000);
#else
    Sleep(25);
#endif
    return;
  }

  tpose.x = unmilli(mage_state[STATE_X]) + origin[0];
  tpose.y = unmilli(mage_state[STATE_Y]) + origin[1];
  tpose.t = unmilli(mage_state[STATE_T]) + origin[2];
  if(trajectory.size() == 0 ||
     (fabs(tpose.x-trajectory.back().x) > 0.01 &&
      fabs(tpose.y-trajectory.back().y) > 0.01))
    trajectory.push_back(tpose);

  // if the server is a simulator, get the true robot pose
  if(server_type == SIMULATOR) {
    if(mpro_get_env_params(&pixpermeter, &origin[0], &origin[1], &origin[2]) < 0) {
      perror("Can't get simulator parameters");
      exit(1);
    }
    if(mpro_get_real_pose(&real_pose.x, &real_pose.y, &real_pose.t) < 0) {
      perror("I can't get the real pose from the simulator!");
      exit(1);
    }
    if((real_trajectory.size() == 0 ||
	(fabs(real_pose.x-real_trajectory.back().x) > 0.01 &&
	 fabs(real_pose.y-real_trajectory.back().y) > 0.01)))
      real_trajectory.push_back(real_pose);
  }

  glutPostRedisplay();
}

int main(int argc, char **argv)
{
  // set up commandline/config file options
  options::add<bool>("help", 0, "Print usage information", 0, false, options::nodump);
  options::add<std::string>("host", "h", "Hostname or IP address of server", 0, "localhost");
  options::add<uint16_t>("port", "p", "Server port to connect to", 0, MPRO_PORT);

  // parse commandline (and read config file if it is specified)
  int inpidx = options::parse_cmdline(argc, argv);
  if(inpidx < 0) // some kind of error
    return 1;    // relevant info printed by getopt

  // print usage information
  if(inpidx > argc || options::quickget<bool>("help") == true) {
    std::cerr << "Usage: " << argv[0] << " [options]" << std::endl;
    options::print_options(std::cout);
    return 1;
  }

  // connect to the server
  fprintf(stderr, "Connecting to %s:%d... ",
	  options::quickget<std::string>("host").c_str(),
	  options::quickget<uint16_t>("port"));
  if(mpro_connect(options::quickget<std::string>("host").c_str(),
                  options::quickget<uint16_t>("port")) < 0) {
    perror("failure!");
    return 1;
  } else
    fprintf(stderr, "success!\n");

  // if the server is a simulator, get the map and environment parameters
  if(server_type == SIMULATOR) {
    fprintf(stderr, "Server is a simulator: requesting the map... ");
    if(mpro_get_map(&map, &map_w, &map_h) < 0) {
      perror("failure!");
      mpro_disconnect();
      return 1;
    }
    fprintf(stderr, "success!\n");
    gui_w = map_w;
    gui_h = map_h;

    if(mpro_get_env_params(&pixpermeter, &origin[0], &origin[1], &origin[2]) < 0) {
      perror("Can't get simulator parameters");
      mpro_disconnect();
      return 1;
    }
    show_map = true;
    show_real_pose = true;
  }

  workspace_w = float(gui_w) / pixpermeter;
  workspace_h = float(gui_h) / pixpermeter;
  orig_gui_w = gui_w;
  orig_gui_h = gui_h;

  // set up glut
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowSize(gui_w, gui_h);
  std::string title = std::string("mprogui: ") +
    options::quickget<std::string>("host") + std::string(" (") +
    std::string((server_type == SIMULATOR ? "simulator)" : "robot)"));
  glutCreateWindow(title.c_str());
  glutDisplayFunc(gui_display);
  glutReshapeFunc(gui_reshape);
  glutKeyboardFunc(gui_keyboard);
  glutSpecialFunc(gui_special);
  glutMouseFunc(gui_mouse);
  glutMotionFunc(gui_motion);
  glutIdleFunc(gui_idle);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_POINT_SMOOTH);
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);

  // start the gui
  glutMainLoop();

  // should never get here
  mpro_disconnect();
  return 0;
}
