/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mproremote.cpp,v 1.2 2006/02/09 01:42:21 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mproremote.cpp
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mproremote.cpp,v 1.2 2006/02/09 01:42:21 beevek Exp $

  \brief A simple remote controller based on libmproclient.  Displays
  an axis.  When the left mouse button is pressed, the position of the
  mouse along the x-axis controls the robot's translational velocity,
  and the position along the y-axis controls the rotational velocity.
 */

#include "mproclient.h"
#include "protocol.h"
#include "options.hpp"
#include <stdio.h>
#include <string>
#include <iostream>
#include <fstream>
#include <GL/glut.h>

// from libmproclient
extern server_type_t server_type;

int W = 400, H = 400;
int mouse_x = 0, mouse_y = 0;
bool on = false;

void update_velocity()
{
  if(on == false)
    mpro_stop();
  else {
    float vt = -2 * MPRO_MAX_TRANS_VEL * (float(mouse_y) / float(H) - 0.5);
    float vr = -2 * MPRO_MAX_ROT_VEL * (float(mouse_x) / float(W) - 0.5);
    mpro_set_velocity(vt, vr);
  }
}

void display()
{
  glClearColor(1, 1, 1, 1);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0, 1, 0, 1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // draw crosshairs
  glColor3f(0, 0, 0);
  glBegin(GL_LINES);
  glVertex2d(0, 0.5);
  glVertex2d(1, 0.5);
  glEnd();
  glBegin(GL_LINES);
  glVertex2d(0.5, 0);
  glVertex2d(0.5, 1);
  glEnd();

  if(on == true) {
    glColor3f(0, 1, 0);
    glBegin(GL_LINES);
    glVertex2d(0.5, 0.5);
    glVertex2d(float(mouse_x) / float(W), float(H-mouse_y) / float(H));
    glEnd();
  }

  glutSwapBuffers();
}

void keyboard(uint8_t key, int x, int y)
{
  if(key == 'q' || key == 'Q') {
    mpro_disconnect();
    exit(0);
  }
}

void mouse(int button, int state, int x, int y)
{
  mouse_x = x;
  mouse_y = y;
  if(button == GLUT_LEFT_BUTTON) {
    on = (state == GLUT_DOWN);
    update_velocity();
  }
  glutPostRedisplay();
}

void motion(int x, int y)
{
  mouse_x = x;
  mouse_y = y;
  if(on == true) {
    update_velocity();
    glutPostRedisplay();
  }
}

int main(int argc, char **argv)
{
  // set up commandline/config file options
  options::add<bool>("help", 0, "Print usage information", 0, false, options::nodump);
  options::add<std::string>("host", "h", "Hostname or IP address of server", 0, "localhost");
  options::add<uint16_t>("port", "p", "Server port to connect to", 0, MPRO_PORT);

  // parse commandline (and read config file if it is specified)
  int inpidx = options::parse_cmdline(argc, argv);
  if(inpidx < 0) // some kind of error
    return 1;    // relevant info printed by getopt

  // print usage information
  if(inpidx > argc || options::quickget<bool>("help") == true) {
    std::cerr << "Usage: " << argv[0] << " [options]" << std::endl;
    options::print_options(std::cout);
    return 1;
  }

  // connect to the server
  fprintf(stderr, "Connecting to %s:%d... ",
	  options::quickget<std::string>("host").c_str(),
	  options::quickget<uint16_t>("port"));
  if(mpro_connect(options::quickget<std::string>("host").c_str(),
                  options::quickget<uint16_t>("port")) < 0) {
    perror("failure!");
    return 1;
  } else
    fprintf(stderr, "success!\n");

  mpro_stop(); // stop the robot

  // set up glut
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowSize(W, H);
  std::string title = std::string("mproremote: ") +
    options::quickget<std::string>("host") + std::string(" (") +
    std::string((server_type == SIMULATOR ? "simulator)" : "robot)"));
  glutCreateWindow(title.c_str());
  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_POINT_SMOOTH);
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);

  // start the gui
  glutMainLoop();

  // should never get here
  mpro_disconnect();
  return 0;
}
