/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mproserver.cpp,v 1.7 2006/02/14 00:52:54 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mproserver.cpp
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mproserver.cpp,v 1.7 2006/02/14 00:52:54 beevek Exp $

  \brief Magellan Pro server library.  Listens for libmproclient
  connections and requests, translates them, and calls the appropriate
  handlers.  Also logs things if a logfile is specified.  A lot of the
  server code is from the "robot" repository in my CVS.
 */

#include "mproserver.h"
#include "protocol.h"
#include <assert.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <list>
#ifndef _WIN32
  #include <unistd.h>
  #include <sys/time.h>
  #include <sys/wait.h>
  #include <sys/socket.h>
  #include <sys/types.h>
  #include <sys/select.h>
  #include <netinet/in.h>
  #include <netinet/tcp.h>
  #include <arpa/inet.h>
  #include <signal.h>
#else
  #include <time.h>
  #include <winsock2.h>
#endif

extern int errno;

// internal use client calls
extern int mpro_get_state(int state[MPRO_STATE_SIZE]);
extern int mpro_get_real_pose(float *x, float *y, float *t);
extern int mpro_get_map(uint8_t **pixbuf, uint32_t *width, uint32_t *height);
extern int mpro_get_env_params(float *pixpermeter,
			       float *origin_x, float *origin_y, float *origin_t);
extern int mpro_poweroff(void);

static socket_t tcp_sock = INVALID_SOCKET;
static server_type_t server_type = CONTROL;
static int (*process_func)(void) = 0;
static uint32_t process_ms = 0;

struct client
{
  struct sockaddr_in ca;
  socket_t sock;
};
std::list<client> clients;

void printtime(void)
{
  static char tstr[100];
  time_t t = time(0);
  struct tm *ltime = localtime(&t);
  strftime(tstr, 100, "[%F %T] ", ltime);
  printf(tstr);
}

int listen_init(uint16_t port)
{
  tcp_sock = socket(AF_INET, SOCK_STREAM, 0);
  if(tcp_sock == INVALID_SOCKET)
    return -1;

  int i = 1;
  if(setsockopt(tcp_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&i, sizeof(i)) < 0 ||
     setsockopt(tcp_sock, IPPROTO_TCP, TCP_NODELAY, (char *)&i, sizeof(i)) < 0) {
    shutdown(tcp_sock,2);
    tcp_sock = INVALID_SOCKET;
    return -1;
  }

  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  sa.sin_port = htons(port);
  if(bind(tcp_sock, (struct sockaddr *)&sa, sizeof(sa)) < 0 ||
     listen(tcp_sock, 5) < 0) {
    shutdown(tcp_sock,2);
    tcp_sock = INVALID_SOCKET;
    return -1;
  }

  return 0;
}

// read a message from the client's socket, parse it, handle it
// appropriately, and send a response if necessary
int handle_message(client &c)
{
  assert(c.sock >= 0);
  static mpro_msg_t msg;
  static int arg[MPRO_STATE_SIZE], tmp, i;
  static float val[MPRO_NUM_IR]; // ALERT: i'm assuming #ir == #sonar
  static bool tog[MPRO_NUM_BUMP];
  static uint8_t *map = 0;
  static uint32_t map_w, map_h;

#define do_read(m) if(sock_recv(c.sock,(char *)&m,sizeof(m)) != sizeof(m)) return -1
#define do_read_n(m) do_read(m); m = ntohl(m)
#define do_write(m) if(send(c.sock,(char *)&m,sizeof(m),0) != sizeof(m)) return -1
#define do_write_n(m) tmp = htonl(m); do_write(tmp);
#define do_write_a(a,n) if(send(c.sock,(char *)a,n*sizeof(uint32_t),0) != n*sizeof(uint32_t)) return -1

  do_read(msg);
  msg.id = ntohl(msg.id);

  switch(msg.cmd) {
  case MSG_POWEROFF:
    printtime();
    printf("Shutting down the robot, adios\n");
    if(mpro_poweroff() < 0)
      perror("I failed to shut down the robot");
    exit(1);
    break;

  case MSG_SET_VELOCITY:
    do_read_n(arg[0]);
    do_read_n(arg[1]);
    mpro_set_velocity(unmilli(arg[0]), unmilli(arg[1]));
    break;

  case MSG_RESET_ODOMETRY:
    do_read_n(arg[0]);
    do_read_n(arg[1]);
    do_read_n(arg[2]);
    mpro_reset_odometry(unmilli(arg[0]), unmilli(arg[1]), unmilli(arg[2]));
    break;

  case MSG_GET_ODOMETRY:
    mpro_get_odometry(&val[0], &val[1], &val[2]);
    arg[0] = milli(val[0]); arg[1] = milli(val[1]); arg[2] = milli(val[2]);
    msg.id = htonl(msg.id);
    do_write(msg);
    do_write_n(arg[0]);
    do_write_n(arg[1]);
    do_write_n(arg[2]);
    break;

  case MSG_GET_ALL_IR:
    mpro_get_all_ir(val);
    msg.id = htonl(msg.id);
    do_write(msg);
    for(i = 0; i < MPRO_NUM_IR; ++i)
      arg[i] = htonl(milli(val[i]));
    do_write_a(arg, MPRO_NUM_IR);
    break;

  case MSG_GET_ALL_SONAR:
    mpro_get_all_sonar(val);
    msg.id = htonl(msg.id);
    do_write(msg);
    for(i = 0; i < MPRO_NUM_SONAR; ++i)
      arg[i] = htonl(milli(val[i]));
    do_write_a(arg, MPRO_NUM_SONAR);
    break;

  case MSG_GET_ALL_BUMP:
    mpro_get_all_bump(tog);
    arg[0] = 0;
    for(i = 0; i < MPRO_NUM_BUMP; ++i)
      arg[0] |= tog[i] << i;
    msg.id = htonl(msg.id);
    do_write(msg);
    do_write_n(arg[0]);
    break;

  case MSG_GET_STATE:
    mpro_get_state(arg);
    msg.id = htonl(msg.id);
    do_write(msg);
    for(i = 0; i < MPRO_STATE_SIZE; ++i)
      arg[i] = htonl(arg[i]);
    do_write_a(arg, MPRO_STATE_SIZE);
    break;

  case MSG_RESET_SIM:
    if(server_type != SIMULATOR) {
      msg.cmd = MSG_ERROR;
      msg.id = htonl(msg.id);
      do_write(msg);
    } else
      mpro_reset_sim();
    break;

  case MSG_GET_REAL_POSE:
    if(server_type != SIMULATOR) {
      msg.cmd = MSG_ERROR;
      msg.id = htonl(msg.id);
      do_write(msg);
    } else {
      mpro_get_real_pose(&val[0], &val[1], &val[2]);
      arg[0] = milli(val[0]); arg[1] = milli(val[1]); arg[2] = milli(val[2]);
      msg.id = htonl(msg.id);
      do_write(msg);
      do_write_n(arg[0]);
      do_write_n(arg[1]);
      do_write_n(arg[2]);
    }
    break;

  case MSG_GET_MAP:
    if(map)
      delete [] map;
    if(server_type != SIMULATOR || mpro_get_map(&map, &map_w, &map_h) < 0) {
      msg.cmd = MSG_ERROR;
      msg.id = htonl(msg.id);
      do_write(msg);
    } else {
      msg.id = htonl(msg.id);
      do_write(msg);
      do_write_n(map_w);
      do_write_n(map_h);
      if(send(c.sock, (char *)map, map_w * map_h, 0) != int(map_w * map_h)) // send the pixels
	return -1;
    }
    break;

  case MSG_GET_ENV_PARAMS:
    if(server_type != SIMULATOR) {
      msg.cmd = MSG_ERROR;
      msg.id = htonl(msg.id);
      do_write(msg);
    } else {
      mpro_get_env_params(&val[0], &val[1], &val[2], &val[3]);
      arg[0] = milli(val[0]); arg[1] = milli(val[1]);
      arg[2] = milli(val[2]); arg[3] = milli(val[3]);
      msg.id = htonl(msg.id);
      do_write(msg);
      do_write_n(arg[0]);
      do_write_n(arg[1]);
      do_write_n(arg[2]);
      do_write_n(arg[3]);
    }
    break;

  case MSG_SET_L_VELOCITY:
  case MSG_SET_R_VELOCITY:
  case MSG_SET_PTZ_ZOOM:
  case MSG_SET_PTZ_PANTILT:
  case MSG_GET_PTZ_IMAGE:
  default:
    // not yet implemented
    msg.cmd = MSG_ERROR;
    msg.id = htonl(msg.id);
    do_write(msg);
    break;

  }

  return 0;
}

int mpro_server_start(server_type_t type, uint16_t port)
{
  server_type = type;
  printtime();
  printf("Initializing server (type 0x%x)...\n", type);
#ifdef _WIN32
  WORD version = MAKEWORD(2,2);
  WSADATA wsa_data;
  if(WSAStartup(version, &wsa_data) != 0)
    return -1;
#else
  signal(SIGPIPE, SIG_IGN);
#endif
  if(listen_init(port) < 0)
    return -1;
  printf("Successfully initialized server\n");
  return 0;
}

int mpro_server_stop()
{
  shutdown(tcp_sock, 2);
  tcp_sock = INVALID_SOCKET;
#ifdef _WIN32
  WSACleanup();
#endif
  return 0;
}

int mpro_server_process_func(int (*process)(void), uint32_t ms)
{
  process_func = process;
  process_ms = ms;
  return 0;
}

int mpro_serve_forever()
{
  assert(tcp_sock != INVALID_SOCKET);
#ifdef _WIN32
  int max = 0; // ignored
#else
  int max = tcp_sock;
#endif
  int ret;
  fd_set rfd;
  struct timeval timeout;
  std::list<std::list<client>::iterator> deleteme;

  while(1) {
    FD_ZERO(&rfd);
    FD_SET(tcp_sock, &rfd);
    for(std::list<client>::iterator i = clients.begin(); i != clients.end(); ++i) {
#ifndef _WIN32
      if(i->sock > max) max = i->sock;
#endif
      FD_SET(i->sock, &rfd);
    }

    if(process_ms > 0) {
      timeout.tv_sec = 0;
      timeout.tv_usec = 1000 * process_ms;
      ret = select(max + 1, &rfd, 0, 0, &timeout);
    } else
      ret = select(max + 1, &rfd, 0, 0, 0);

    if(ret < 0) { // an error
      if(errno == EINTR)
	continue;
      perror("select()");
      return -1;
    }

    if(process_func != 0 && process_func() < 0)
      return -1;

    if(ret == 0) // no data, just the processing timeout
      continue;

    // data is waiting

    if(FD_ISSET(tcp_sock, &rfd)) {
      struct client c;
#ifdef _WIN32
      int calen = sizeof(c.ca);
#else
      socklen_t calen = sizeof(c.ca);
#endif
      c.sock = accept(tcp_sock, (struct sockaddr *)&(c.ca), &calen);
      if(c.sock < 0) {
	perror("accept()");
	return -1;
      }
      printtime(); printf("Connection from %s\n", inet_ntoa(c.ca.sin_addr));
      clients.push_back(c);

      // send server type
      mpro_msg_t msg;
      msg.id = 0;
      msg.cmd = MSG_SERVER_TYPE;
      if(send(c.sock, (char *)&msg, sizeof(msg), 0) != sizeof(msg)) return -1;
      if(send(c.sock, (char *)&server_type, sizeof(server_type), 0) != sizeof(server_type)) return -1;
    }

    for(std::list<client>::iterator i = clients.begin(); i != clients.end(); ++i)
      if(FD_ISSET(i->sock, &rfd) && handle_message(*i) < 0) {
	shutdown(i->sock, 2);
	printtime(); printf("Client %s disconnected... ", inet_ntoa(i->ca.sin_addr));
	mpro_stop(); printf("stopped\n"); // stop whenever there's a disconnection
	deleteme.push_back(i);
      }

    std::list<std::list<client>::iterator>::iterator j = deleteme.begin();
    for(; j != deleteme.end(); ++j)
      clients.erase(*j);
    deleteme.clear();

    if(clients.size() == 0)
      mpro_stop();
  }
}
