/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: mprosim.cpp,v 1.5 2006/02/09 01:42:21 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file mprosim.cpp
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: mprosim.cpp,v 1.5 2006/02/09 01:42:21 beevek Exp $

  \brief Magellan Pro simulator.  Simulates robot kinematics at a
  fairly high frequency and listens for connections from libmproclient
  to update the robot's state.  Note that the GUI is a separate
  process that connects to the simulator as a client.
 */

#include "mproserver.h"
#include "options.hpp"
#include <string>
#include <iostream>
#include <fstream>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#ifndef _WIN32
  #include <unistd.h>
  #include <sys/time.h>
#else
  #define WIN32_LEAN_AND_MEAN
  #include <windows.h>
  #define M_PI 3.1415926535897932384626433833
#endif

#ifdef _WIN32
void srand48(long seed) { srand(seed); }
double drand48()
{
  return double(rand())/double(RAND_MAX + 1);
}
int gettimeofday (struct timeval *tv, void* tz)
{
  union {
    long long ns100; /* time since 1 Jan 1601 in 100ns units */
    FILETIME ft;
  } now;
  GetSystemTimeAsFileTime (&now.ft);
  tv->tv_usec = (long) ((now.ns100 / 10LL) % 1000000LL);
  tv->tv_sec = (long) ((now.ns100 - 116444736000000000LL) / 10000000LL);
  return (0);
}
#endif

extern int errno;

// from MAGE:
#define STATE_IR_0                 1
#define STATE_SONAR_0              17 
#define STATE_BUMPER               33
#define STATE_X                    34
#define STATE_Y                    35
#define STATE_T                    36
#define STATE_VEL_TRANS            37
#define STATE_VEL_STEER            38

// in pgm.cpp
extern bool load_pgm(const char *file, uint8_t **grid, uint32_t *w, uint32_t *h);
int mpro_reset_sim(void); // declared below

#define SIM_FREQUENCY_MS 40  // milliseconds between simulation updates
struct timeval now, then;    // don't trust the requested time: measure instead!

// robot state
struct state_t {
  float pose[3];                // real pose (in world coordinates)
  float origin[3];              // odometry origin
  float odom[3];                // odometry pose
  float v_trans, v_rot;         // translational/rotational velocities
  float ir[MPRO_NUM_IR];        // ir ranges (only updated when requested)
  float sonar[MPRO_NUM_SONAR];  // sonar ranges (only updated when requested)
  bool  bump[MPRO_NUM_BUMP];    // bump toggles
} state;

// sensor noise models (proportional to range)
float sigma_ir, sigma_sonar;
// sonar reflection coefficients: a reflection occurs with probability
//  P(r) = max(1, scale * (abs(angle_beam - angle_wall) - 90)/90 + base)
float reflect_scale, reflect_base;
// motion noise models (proportional to distance)
float sigma_trans, sigma_rot, sigma_normal;

// grid-based map
// world coordinates are centered at the center of the map grid
struct map_t {
  uint8_t *grid;                // map pixels
  uint32_t w, h;                // grid width, height
  float ppm;                    // map pixels per meter
  float xmin, xmax, ymin, ymax; // map bounding box (world coords)

  // get a reference to the pixel at (r,c)
  inline uint8_t & pixel(uint32_t r, uint32_t c)
  {
    assert(r < w && c < h);
    return grid[r * w + c];
  }

  inline void coords(uint32_t r, uint32_t c, float *x, float *y)
  {
    assert(r < w && c < h);
    *x = c / ppm + xmin;
    *y = r / ppm + ymin;
  }

  inline bool occupied(uint32_t r, uint32_t c)
  {
    if(r >= w || c >= h)
      return false;
    return pixel(r, c) < 127;
  }

  // get the row and column in the grid where (x,y) (in world
  // coordinates) falls
  inline void cell(float x, float y, uint32_t *r, uint32_t *c)
  {
    assert(r && c);
    *c = (uint32_t)floor((x - xmin) * ppm);
    *r = (uint32_t)floor((y - ymin) * ppm);
  }
} map;


/*
  Grid map utilities
*/

// returns true if intersection, false otherweise; puts distance along
// line from (x0,y0) of intersection with the map in range (0 if no
// intersection)
bool intersect_map_line(float x0, float y0, float x1, float y1, float *range)
{
  uint32_t r, c;
  float dx = x1 - x0, dy = y1 - y0;
  uint32_t count = 2 * (uint32_t)floor(sqrt(dx*dx + dy*dy) * map.ppm) + 1;
  dx /= float(count); // step size (x)
  dy /= float(count); // step size (y)
  for(x1 = x0, y1 = y0; count > 0; --count) { // step along the line
    map.cell(x1, y1, &r, &c);
    if(r >= map.h || c >= map.w)
      continue; // outside the grid
    if(map.occupied(r, c)) {
      // FIXME: we should actually find the closest point in the pixel
      // to (x0,y0) and return that range instead
      *range = sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0));
      return true;
    }
    x1 += dx;
    y1 += dy;
  }
  *range = 0;
  return false;
}

// look at pixels in the 9x9 neighborhood of (r,c) and fit a line to
// the center points of the ones that are occupied
void local_line_approximation(uint32_t r, uint32_t c,
			      float *x0, float *y0, float *x1, float *y1)
{
  uint32_t i, j, N = 0;
  float x[81], y[81];

  // get data points for the centers of occupied cells
  for(i = r - 4; i <= r + 4; ++i)
    for(j = c - 4; j <= c + 4; ++j)
      if(map.occupied(i, j)) {
	map.coords(i, j, &x[N], &y[N]);
	++N;
      }

  // linear least squares (based loosely on GSL's version)
  float ux = 0, uy = 0, udx2 = 0, udxdy = 0, dx, dy;
  for(i = 0; i < N; ++i) {
    ux += (x[i] - ux) / (i + 1.0);
    uy += (y[i] - uy) / (i + 1.0);
  }
  for(i = 0; i < N; ++i) {
    dx = x[i] - ux;
    dy = y[i] - uy;
    udx2 += (dx * dx - udx2) / (i + 1.0);
    udxdy += (dx * dy - udxdy) / (i + 1.0);
  }
  float m = udxdy / udx2; // slope
  float b = uy - ux * m;  // intersect

  // return the line in two-point form
  *x0 = 0.0;
  *y0 = b;
  *x1 = 1.0;
  *y1 = m + b;
}

/*
  Simulation 
*/

double gaussian(double mean, double stddev)
{
  static bool cached = false;
  static double extra;

  if(cached) {
    cached = false;
    return extra * stddev + mean;
  }
  
  // pick random point in unit circle
  double a, b, c;
  do {
    a = 2.0 * drand48() - 1.0;
    b = 2.0 * drand48() - 1.0;
    c = a * a + b * b;
  } while(c >= 1.0 || c == 0.0);
  double t = double(::sqrt(-2.0 * ::log(c) / c));
  extra = t * a;
  cached = true;
  return t * b * stddev + mean;
}

int sim_motion(void)
{
  // get elapsed time since last update
  gettimeofday(&now, 0);
  double dt = double(now.tv_sec) + double(now.tv_usec)/1000000.0 -
    double(then.tv_sec) - double(then.tv_usec)/1000000.0;

  // do kinematics simulation

  // FIXME: this is a bad model: we should be using the translational
  // velocity to compute an arclength along a circular path; instead
  // we'll just translate then rotate for now.
  float translate = state.v_trans * dt;
  float rotate = state.v_rot * dt;

  // update real pose
  state.pose[0] += translate * cos(state.pose[2]);
  state.pose[1] += translate * sin(state.pose[2]);
  state.pose[2] += rotate;

  // update odometry (add noise)
  float noisy_translate = gaussian(translate, sigma_trans*translate);
  state.odom[0] += noisy_translate * cos(state.odom[2] + state.origin[2]);
  state.odom[1] += noisy_translate * sin(state.odom[2] + state.origin[2]);
  state.odom[2] += gaussian(rotate, sigma_rot*rotate + sigma_normal*translate);

  // if we run into something, stop (and update bump sensors)

  // first check all the pixels inside the robot
  uint32_t r, c, rad = (uint32_t)ceil(MPRO_RADIUS * map.ppm);
  float x, y;
  map.cell(state.pose[0], state.pose[1], &r, &c);
  for(uint32_t i = r - rad; i <= r + rad; ++i)
    for(uint32_t j = c - rad; j <= c + rad; ++j)
      if(map.occupied(i, j)) {
	map.coords(i, j, &x, &y);
	if(sqrt((state.pose[0]-x)*(state.pose[0]-x) + (state.pose[1]-y)*(state.pose[1]-y)) < MPRO_RADIUS) {
	  mpro_stop();
	  break;
	}
      }

  // then check the perimeter of the robot (fine)
  float a = -M_PI/16.0, x0, y0, x1, y1, range;
  for(uint32_t i = 0; i < MPRO_NUM_BUMP; ++i, a += M_PI/8.0) {
    x0 = state.pose[0] + MPRO_RADIUS * cos(a+state.pose[2]);
    y0 = state.pose[1] + MPRO_RADIUS * sin(a+state.pose[2]);
    x1 = state.pose[0] + MPRO_RADIUS * cos(a+M_PI/8.0+state.pose[2]);
    y1 = state.pose[1] + MPRO_RADIUS * sin(a+M_PI/8.0+state.pose[2]);
    if(intersect_map_line(x0, y0, x1, y1, &range)) {
      mpro_stop();
      state.bump[i] = true;
    } else
      state.bump[i] = false;
  }

  then = now;
  return 0;
}

void sim_update_ir(void)
{
  // raytrace in the map to get ir ranges
  float a, x0, y0, x1, y1;
  for(int i = 0; i < MPRO_NUM_IR; ++i) {
    a = float(i) * 2.0 * M_PI/float(MPRO_NUM_IR) + state.pose[2];
    x0 = state.pose[0] + MPRO_RADIUS * cos(a);
    y0 = state.pose[1] + MPRO_RADIUS * sin(a);
    x1 = x0 + MPRO_IR_MAX * cos(a);
    y1 = y0 + MPRO_IR_MAX * sin(a);
    if(intersect_map_line(x0, y0, x1, y1, &state.ir[i]) &&
       (state.ir[i] >= MPRO_IR_MIN || state.ir[i] <= MPRO_IR_MAX))
      state.ir[i] += gaussian(0.0, sigma_ir * state.ir[i]); // add noise
    else
      state.ir[i] = MPRO_IR_MAX + 0.01; // out of range
  }
}

// pretend the origin of a sonar beam is at (x0,y0,a) and the max
// radius is r; model it.  recursively calls itself for reflections.
float *samples = 0;
uint32_t N = 0;
float model_sonar(float x0, float y0, float a, float range)
{
  // compute the number of samples we need to miss no pixels at the
  // maximum range
  if(!samples) {
    N = (uint32_t)(MPRO_SONAR_MAX * tan(MPRO_SONAR_BEAM_WIDTH/2.0) * map.ppm);
    N /= 3; // FIXME: remove this at some point; better yet, find a
	    // more efficient way than raytracing to find the closest
	    // occupied pixel; this is the cause of framerate problems
    samples = new float[N];
  }
  float cura = a - MPRO_SONAR_BEAM_WIDTH/2.0, ainc = MPRO_SONAR_BEAM_WIDTH/float(N-1);
  float x1, y1;
  uint32_t minint = N;

  // first compute ranges for all of the samples
  for(uint32_t i = 0; i < N; ++i, cura += ainc) {
    x1 = x0 + range * cos(cura);
    y1 = y0 + range * sin(cura);
    if(!intersect_map_line(x0, y0, x1, y1, &samples[i]))
      samples[i] = range + 0.01; // out of range
    else if(samples[i] > MPRO_SONAR_MIN && (minint == N || samples[i] < samples[minint]))
      minint = i; // closest so far
  }

  if(minint == N) // no intersections
    return samples[0];

  // FIXME: remove after fixing the reflection model
  return samples[minint];

  // for the closest sample, find the angle of the beam with wall hit
  // by that sample and determine whether a reflection occurs
  cura = a - MPRO_SONAR_BEAM_WIDTH/2.0 + ainc * float(minint);
  x1 = x0 + samples[minint] * cos(cura);
  y1 = y0 + samples[minint] * sin(cura);
  uint32_t r, c;
  map.cell(x1, y1, &r, &c);
  float wx0, wy0, wx1, wy1;
  local_line_approximation(r, c, &wx0, &wy0, &wx1, &wy1);
  // FIXME:
  //  compute angle between the two lines
  //  use this and the reflection model to decide if a reflection occurs
  //  if so, compute the new "origin," angle and the remaining range,
  //   recursively call model_sonar, and add the returned range to the
  //   distance to the reflection point
  return range + 0.01; // FIXME remove
}

void sim_update_sonar(void)
{
  // raytrace in the map to get sonar ranges
  float a, x0, y0;
  for(int i = 0; i < MPRO_NUM_SONAR; ++i) {
    a = float(i) * 2.0 * M_PI/float(MPRO_NUM_SONAR) + state.pose[2];
    x0 = state.pose[0] + MPRO_RADIUS * cos(a);
    y0 = state.pose[1] + MPRO_RADIUS * sin(a);
    state.sonar[i] = model_sonar(x0, y0, a, MPRO_SONAR_MAX);
    if(state.sonar[i] < MPRO_SONAR_MIN || state.sonar[i] > MPRO_SONAR_MAX)
      state.sonar[i] = MPRO_SONAR_MAX + 0.01; // out of range
    else
      state.sonar[i] += gaussian(0.0, sigma_sonar * state.sonar[i]); // add noise
  }
}


int main(int argc, char **argv)
{
  // set up commandline/config file options
  options::add<bool>("help", 0, "Print usage information", 0, false, options::nodump);
  options::set_cf_options("config", "c");
  options::add<std::string>("save-config", 0, "Save configuration file", 0, "", options::nodump);

  options::add<std::string>("map", "m", "Map file (PGM format) [REQUIRED]", "Environment", "");
  options::add<float>("pixels-per-meter", "p", "Pixels per meter", "Environment", 20.0);

  options::add<float>("origin-x", "x", "Robot origin (x)", "Robot", 0.0);
  options::add<float>("origin-y", "y", "Robot origin (y)", "Robot", 0.0);
  options::add<float>("origin-theta", "t", "Robot origin (theta)", "Robot", 0.0);

  options::add<float>("err-ir", 0, "IR range error (% of range measurement)",
		      "Error models", 0.07);
  options::add<float>("err-sonar", 0, "Sonar range error (% of range measurement)",
		      "Error models", 0.07);
  options::add<float>("err-reflection-base", 0, "Sonar reflection base probability",
		      "Error models", 0.08);
  options::add<float>("err-reflection-scale", 0, "Sonar reflection scaling factor (probability increase with angle)",
		      "Error models", 0.8);
  options::add<float>("err-translation", 0, "Translation error (% of odometry measurement)",
		      "Error models", 0.1);
  options::add<float>("err-rotation", 0, "Rotation error (% of odometry measurement)",
		      "Error models", 0.3);
  options::add<float>("err-normal", 0, "Normal error (% of odometry measurement)",
		      "Error models", 0.1);

  // parse commandline (and read config file if it is specified)
  int inpidx = options::parse_cmdline(argc, argv);
  if(inpidx < 0) // some kind of error
	return 1;    // relevant info printed by getopt

  std::string mapfile = options::quickget<std::string>("map");

  // print usage information
  if(inpidx > argc || options::quickget<bool>("help") == true || mapfile == "") {
	std::cerr << "Usage: " << argv[0] << " [options]" << std::endl;
    options::print_options(std::cout);
    return 1;
  }

  // save a config file based on these options?
  std::string cfname = options::quickget<std::string>("save-config");
  if(cfname.length() > 0) {
    std::ofstream conf(cfname.c_str());
    if(!conf)
      std::cerr << "Can't write configuration file " << cfname << std::endl;
    else
      options::dump(conf);
  }

  // load map
  if(!load_pgm(mapfile.c_str(), &map.grid, &map.w, &map.h)) {
    std::cerr << "Can't load map file " << mapfile << std::endl;
    return 1;
  }
  map.ppm = options::quickget<float>("pixels-per-meter");
  map.xmin = -float(map.w) / 2 / map.ppm;
  map.xmax = -map.xmin;
  map.ymin = -float(map.h) / 2 / map.ppm;
  map.ymax = -map.ymin;

  mpro_reset_sim(); // initial pose

  // set error models
  sigma_ir = options::quickget<float>("err-ir");
  sigma_sonar = options::quickget<float>("err-sonar");
  reflect_scale = options::quickget<float>("err-reflection-scale");
  reflect_base = options::quickget<float>("err-reflection-base");
  sigma_trans = options::quickget<float>("err-translation");
  sigma_rot = options::quickget<float>("err-rotation");
  sigma_normal = options::quickget<float>("err-normal");

  srand48(time(0));
  gettimeofday(&then, 0);

  // start the server
  mpro_server_process_func(sim_motion, SIM_FREQUENCY_MS);
  if(mpro_server_start(SIMULATOR) < 0 || mpro_serve_forever() < 0)
    perror("ERROR");

  mpro_server_stop(); // should never get here
  return 0;
}

/*
  Implementations of client request handlers.
*/
int mpro_set_velocity(float trans, float rot)
{
  state.v_trans = trans;
  state.v_rot = rot;
  return 0;
}

int mpro_stop(void)
{
  return mpro_set_velocity(0, 0);
}

// only on the simulator
int mpro_reset_sim(void)
{
  // set initial pose
  state.pose[0] = state.origin[0] = options::quickget<float>("origin-x");
  state.pose[1] = state.origin[1] = options::quickget<float>("origin-y");
  state.pose[2] = state.origin[2] = options::quickget<float>("origin-theta");
  state.odom[0] = state.odom[1] = state.odom[2] = 0;
  return 0;
}

int mpro_reset_odometry(float x, float y, float t)
{
  state.origin[0] = x;
  state.origin[1] = y;
  state.origin[2] = t;
  state.odom[0] = state.odom[1] = state.odom[2] = 0.0;
  return 0;
}

int mpro_get_odometry(float *x, float *y, float *t)
{
  assert(x != 0 && y != 0 && t != 0);
  *x = state.odom[0];
  *y = state.odom[1];
  *t = state.odom[2];
  return 0;
}

int mpro_get_all_ir(float ranges[MPRO_NUM_IR])
{
  sim_update_ir();
  memcpy(ranges, state.ir, MPRO_NUM_IR * sizeof(float));
  return 0;
}

int mpro_get_ir(uint8_t index, float *range)
{
  assert(index < MPRO_NUM_IR && range != 0);
  sim_update_ir();
  *range = state.ir[index];
  return 0;
}

int mpro_get_all_sonar(float ranges[MPRO_NUM_SONAR])
{
  sim_update_sonar();
  memcpy(ranges, state.sonar, MPRO_NUM_SONAR * sizeof(float));
  return 0;
}

int mpro_get_sonar(uint8_t index, float *range)
{
  assert(index < MPRO_NUM_SONAR && range != 0);
  sim_update_sonar();
  *range = state.sonar[index];
  return 0;
}

int mpro_get_all_bump(bool toggles[MPRO_NUM_BUMP])
{
  memcpy(toggles, state.bump, MPRO_NUM_BUMP * sizeof(bool));
  return 0;
}

int mpro_get_bump(uint8_t index, bool *toggle)
{
  assert(index < MPRO_NUM_BUMP && toggle != 0);
  *toggle = state.bump[index];
  return 0;
}

int mpro_get_state(int mage_state[MPRO_STATE_SIZE])
{
  // contruct a mage state from the simulated data
  memset(mage_state, 0, MPRO_STATE_SIZE * sizeof(int));
  mage_state[STATE_X] = milli(state.odom[0]);
  mage_state[STATE_Y] = milli(state.odom[1]);
  mage_state[STATE_T] = milli(state.odom[2]);
  mage_state[STATE_VEL_TRANS] = milli(state.v_trans);
  mage_state[STATE_VEL_STEER] = milli(state.v_rot);
  sim_update_ir();
  sim_update_sonar();
  int i;
  for(i = 0; i < MPRO_NUM_IR; ++i)
    mage_state[STATE_IR_0+i] = milli(state.ir[i]);
  for(i = 0; i < MPRO_NUM_SONAR; ++i)
    mage_state[STATE_SONAR_0+i] = milli(state.sonar[i]);
  for(i = 0; i < MPRO_NUM_BUMP; ++i)
    if(state.bump[i])
      mage_state[STATE_BUMPER] |= 0x1 << i;
  return 0;
}

// only on the simulator
int mpro_get_real_pose(float *x, float *y, float *t)
{
  assert(x && y && t);
  *x = state.pose[0];
  *y = state.pose[1];
  *t = state.pose[2];
  return 0;
}

int mpro_get_map(uint8_t **pixbuf, uint32_t *width, uint32_t *height)
{
  assert(pixbuf && width && height);
  if(map.grid == 0) {
    errno = ENOENT;
    return -1;
  }
  *pixbuf = new uint8_t[map.w * map.h];
  memcpy(*pixbuf, map.grid, map.w * map.h);
  *width = map.w;
  *height = map.h;
  return 0;
}

int mpro_get_env_params(float *pixpermeter,
			float *origin_x, float *origin_y, float *origin_t)
{
  assert(pixpermeter && origin_x && origin_y && origin_t);
  *pixpermeter = map.ppm;
  *origin_x = state.origin[0];
  *origin_y = state.origin[1];
  *origin_t = state.origin[2];
  return 0;
}

// only on the real robot
int mpro_poweroff(void)
{
  return 0;
}

// not yet implemented
int mpro_set_l_velocity(float v) { return -1; }
int mpro_set_r_velocity(float v) { return -1; }
int mpro_set_ptz_zoom(float zoom) { return -1; }
int mpro_set_ptz_pantilt(float pan, float tilt) { return -1; }
int mpro_get_ptz_image(uint8_t **pixbuf, uint32_t *width, uint32_t *height, uint8_t *bpp) { return -1; }
