/*
  This file is part of mpro.

  mpro is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: protocol.h,v 1.4 2006/01/30 05:38:43 beevek Exp $

  Copyright (c) 2006 Kris Beevers
*/

/*!
  \file protocol.h
  \author Kris Beevers (beevek@cs.rpi.edu)
  \version $Id: protocol.h,v 1.4 2006/01/30 05:38:43 beevek Exp $

  \brief Network protocol (used by both clients and servers)
 */

#ifndef _PROTOCOL_H
#define _PROTOCOL_H

#ifdef __cplusplus
extern "C" {
#endif

#include <inttypes.h>
#include <sys/types.h>
#ifndef _WIN32
  typedef int socket_t;
  #define INVALID_SOCKET -1
#else
  #include <winsock2.h>
  typedef SOCKET socket_t;
#endif

#define MPRO_STATE_SIZE 46 // in 4-byte blocks

enum {
  MSG_SET_VELOCITY     = 0x01,
  MSG_SET_L_VELOCITY   = 0x02,
  MSG_SET_R_VELOCITY   = 0x03,

  MSG_RESET_ODOMETRY   = 0x20,
  MSG_GET_ODOMETRY     = 0x21,
  MSG_GET_ALL_IR       = 0x22,
  MSG_GET_ALL_SONAR    = 0x23,
  MSG_GET_ALL_BUMP     = 0x24,

  MSG_SET_PTZ_ZOOM     = 0x30,
  MSG_SET_PTZ_PANTILT  = 0x31,
  MSG_GET_PTZ_IMAGE    = 0x32,

  MSG_ERROR            = 0x00,

  // extended commands

  MSG_SERVER_TYPE      = 0x40, // what kind of server (simulator, controller...)
  MSG_GET_STATE        = 0x41, // get the full robot state vector (minus camera)
  MSG_GET_MAP          = 0x42, // get the simulator's map (only in the simulator)
  MSG_GET_REAL_POSE    = 0x43, // get the robot's real pose (only in the simulator)
  MSG_GET_ENV_PARAMS   = 0x44, // get simulator parameters (origin, pixels per meter)
  MSG_RESET_SIM        = 0x45, // reset the simulation to the initial state
  MSG_POWEROFF         = 0xff, // power down the robot (only on the real robot)
};

typedef enum {
  CONTROL       = 0x00,
  SIMULATOR     = 0x01
} server_type_t;

/*
  request format:
    request id: 4 bytes
    command: 1 byte
    arguments: 4 bytes each

  response format:
    request id: 4 bytes
    command (from request): 1 byte
    data: per command (generally 4 bytes each)
*/
typedef struct
{
  uint32_t id;  // not really used yet but potentially useful
  uint8_t  cmd;
} mpro_msg_t;

// convert a real decimal value into an integral value in units of
// 1/1000th (e.g. m->mm, rad->mrad), and vice versa
inline int milli(float f) { return int(1000.0 * f); }
inline float unmilli(int i) { return float(i) / 1000.0; }

// since win32 doesn't support MSG_WAITALL, we'll just do it ourselves
int sock_recv(socket_t sock, char *buf, int size);

#ifdef __cplusplus
}
#endif

#endif // _PROTOCOL_H
